export function addQueryParamsToUrl(url, queryParams = {}) {
    const parsedUrl = new URL(url, globalThis.location.origin);
    url = url.replace(parsedUrl.search, "");
    let search = parsedUrl.searchParams;
    for (const [key, values] of Object.entries(queryParams)) {
        if (values !== undefined) {
            [].concat(values).forEach((value) => search.append(key, value));
        }
    }
    let qs = search.toString();
    return qs.length === 0 ? url : `${url}?${qs}`;
}
function createRequestInit(method, body) {
    let init = { method };
    if (body?.constructor?.name !== "Object") {
        init.body = body;
        return init;
    }
    init.headers = { "content-type": "application/json" };
    if (body) {
        init.body = JSON.stringify(body);
    }
    return init;
}
export function createApiClient(factory, safe) {
    const { visitor, credentials } = factory;
    async function unwrap(cb) {
        try {
            let ok = await cb();
            return safe ? [ok, null] : ok;
        }
        catch (cause) {
            let error = new Error(cause.message, { cause });
            error.name = "CallFailed";
            if (safe) {
                return [null, error];
            }
            else {
                throw error;
            }
        }
    }
    async function call(method, url, qs = {}, body) {
        const init = createRequestInit(method, body);
        const fullUrl = addQueryParamsToUrl(url, qs);
        let updatedInit = await credentials.attach({ ...init }, fullUrl);
        let request = new Request(fullUrl, updatedInit);
        let response = await fetch(request);
        if (response.ok) {
            return unwrap(() => visitor.ok(response));
        }
        let isValid = await credentials.isValid(request, response);
        if (!isValid) {
            let isComplete = await credentials.applyRefresh();
            if (!isComplete) {
                let err = new Error("RefreshFailed");
                if (safe) {
                    return [null, err];
                }
                throw err;
            }
            let updatedInit = await credentials.attach({ ...init }, fullUrl);
            response = await fetch(new Request(fullUrl, updatedInit));
            if (response.ok) {
                return unwrap(() => visitor.ok(response));
            }
        }
        return unwrap(() => visitor.fail(request, response));
    }
    return {
        call,
        get: (url, qs, body) => call("GET", url, qs, body),
        post: (url, qs, body) => call("POST", url, qs, body),
        delete: (url, qs, body) => call("DELETE", url, qs, body),
        postBody: (url, body) => call("POST", url, {}, body),
        put: (url, body) => call("PUT", url, {}, body),
    };
}
